-- taskmgr.e by Mike Nelson
-- Prioritized Cooperative Multitasking
-- Adapted from Method Euphoria Task Mangager Library.
-- Timing code adapted from RDS sched.e in the LANGWAR demo. 

without trace
without warning

include machine.e -- for tick_rate()
include misc.e  -- for platform() constants
       

if platform()=DOS32 then
   tick_rate(100)   
end if 

constant DAY_SECONDS=24*60*60   
integer day_inc day_inc=0
atom last_time last_time=0.0

global function get_time()
   atom t  
   if platform()=DOS32 then
      t=time()+day_inc
      if t<last_time then 
         day_inc+=DAY_SECONDS
         t+=DAY_SECONDS 
      end if
      last_time=t   
      return t 
   else
      return time()
   end if                                      
end function                                  

sequence task_list task_list={}
sequence task_priority task_priority={}
sequence task_activate task_activate={} 
sequence task_repeat task_repeat={} 
sequence task_params task_params={}  
sequence task_message task_message={}   
integer this_task this_task=0 
integer task_pointer task_pointer=0
integer priority_step priority_step=10
integer waiting waiting=0
sequence p_waiting p_waiting=repeat({},11)
integer p_step p_step=11

global constant TASK_INACTIVE=-1.0 
global constant TASK_NOW=0.0
global constant TASK_ONCE=-1.0 
global constant TASK_NO_WAIT=0.0
global constant TASK_MAX_INTERVAL=DAY_SECONDS
global constant TASK_MAX_PRIORITY=10
global constant TASK_NORMAL_PRIORITY=5
global constant TASK_MIN_PRIORITY=0

type valid_priority(object x)
   if integer(x) then
      return x>=TASK_MIN_PRIORITY and x<=TASK_MAX_PRIORITY   
   end if
   return 0
end type   

type nonneg_int(object x)
   if integer(x) then 
      return x>=0
   else
      return 0
   end if
end type                   

global type task(object x)
   if integer(x) then
      return x>0 and x<=length(task_list)   
   end if
end type
                      
global function get_priority(task t)
   return task_priority[t]   
end function                  

global procedure set_priority(task t,valid_priority priority)
   task_priority[t]=priority
end procedure 

global function get_activate(task t)
   return task_activate[t]   
end function                  

global procedure set_activate(task t,atom activate)
   if activate<TASK_NOW or activate>TASK_MAX_INTERVAL then 
      activate=TASK_INACTIVE  
   else
      activate+=get_time()
      activate=floor(activate*100+.5)/100
   end if  
   task_activate[t]=activate 
end procedure 

global function get_repeat(task t)
   return task_repeat[t]   
end function                  

global procedure set_repeat(task t,atom reptime)
   if reptime<TASK_NO_WAIT or reptime>TASK_MAX_INTERVAL then 
      reptime=TASK_ONCE
   else
      reptime=floor(reptime*100+.5)/100
   end if  
   task_repeat[t]=reptime 
end procedure 

global function get_parameters(task t)
   return task_params[t]   
end function                  

global function get_message(task t)
   return task_message[t]   
end function                  
 
global procedure set_message(task t,object message)
   task_message[t]=message 
end procedure 

global procedure reschedule(object t,object priority,atom activate,
atom reptime,sequence params) 
   set_priority(t,priority) 
   set_activate(t,activate)
   set_repeat(t,reptime)
   task_params[t]=params
end procedure   

global function new_task(nonneg_int ID,object priority,atom activate,
atom reptime,sequence params)
   integer t 
   task_list&=ID
   task_priority&=0 
   task_activate&=0
   task_repeat&=0
   task_params&=0    
   task_message&=0 
   t=length(task_list)
   reschedule(t,priority,activate,reptime,params)                   
   return t
end function

global procedure next_task()  
   integer priority
   atom activate,now         
   sequence pw
   object result   
   if this_task!=0 then return end if
   if waiting=0 then  
      now=get_time()
      p_step-=1
      if p_step<0 then p_step=10 end if
      for i=1 to length(task_list) do 
         priority=task_priority[i]
         if priority>=p_step then
            activate=task_activate[i]
            if activate!=TASK_INACTIVE and activate<=now then
               waiting+=1 
               p_waiting[11-task_priority[i]]&=i      
            end if
         end if
      end for
   end if
   if waiting=0 then return end if
   for i=1 to 11 do
      pw=p_waiting[i]
      if length(pw) then
         this_task=pw[1]
         p_waiting[i]=pw[2..$] 
         exit  
      end if
   end for
   waiting-=1                 
   result=call_func(task_list[this_task],task_params[this_task])  
   if atom(result) then result={result} end if
   task_params[this_task]=result
   if task_repeat[this_task]=TASK_ONCE then 
      task_activate[this_task]=TASK_INACTIVE
   end if  
   if task_activate[this_task]!=TASK_INACTIVE then       
      task_activate[this_task]+=task_repeat[this_task] 
   end if 
   this_task=0 
end procedure

global function current_task()
   return this_task   
end function 


